import logging
import os
import math
import glob

import pyfastaq
import pysam
import pandas as pd

from Bio import pairwise2, SeqIO

from cluster_vcf_records import vcf_clusterer, vcf_file_read

from minos import dependencies, dnadiff, plots, utils

class Error (Exception): pass

class DnadiffMappingBasedVerifier:
    '''dnadiff_snps_file = file of snp calls generated by dnadiff.
    dnadiff_file1 = file containing reference sequence passed to dnadiff.
    dnadiff_file2 = file containing reference sequence passed to dnadiff.
    vcf_file1_in = input VCF file corresponding to dnadiff reference sequence, to be verified.
    vcf_file2_in = input VCF file corresponding to dnadiff query, to be verified.
    vcf_reference_file = reference sequence file that was used to make vcf_file_in.
    outprefix = prefix of output files.
    flank_length = length of "truth" sequence to take before/after alleles when mapping.

    Writes 3 files:
    outprefix.1.sam = mapping of seqs + flanks from dnadiff reference sequence to seqs + flanks for first vcf file.
    outprefix.2.sam = mapping of seqs + flanks from dnadiff query sequence to seqs + flanks for second vcf file.
    outprefix.vcf = VCF file with annotations for validation
    outprefix.stats.tsv = summary stats (see dict output by
                          _parse_sam_file_and_update_vcf_records_and_gather_stats()
                          for a description)'''
    def __init__(self, dnadiff_snps_file, dnadiff_file1, dnadiff_file2, vcf_file_in1, vcf_file_in2, vcf_reference_file, outprefix, flank_length=31, merge_length=None, filter_and_cluster_vcf=True, discard_ref_calls=True, allow_flank_mismatches=True, exclude_regions_bed_file1=None, exclude_regions_bed_file2=None, max_soft_clipped=3):
        self.dnadiff_snps_file = os.path.abspath(dnadiff_snps_file)
        self.dnadiff_file1 = os.path.abspath(dnadiff_file1)
        self.dnadiff_file2 = os.path.abspath(dnadiff_file2)
        self.vcf_file_in1 = os.path.abspath(vcf_file_in1)
        self.vcf_file_in2 = os.path.abspath(vcf_file_in2)
        self.vcf_reference_file = os.path.abspath(vcf_reference_file)
        self.sam_file_out1 = os.path.abspath(outprefix + '.1.sam')
        self.sam_file_out2 = os.path.abspath(outprefix + '.2.sam')
        self.seqs_out_dnadiff1 = os.path.abspath(outprefix + '.dnadiff1.fa')
        self.seqs_out_dnadiff2 = os.path.abspath(outprefix + '.dnadiff2.fa')
        self.filtered_vcf1 = os.path.abspath(outprefix + '.1.filter.vcf')
        self.filtered_vcf2 = os.path.abspath(outprefix + '.2.filter.vcf')
        self.clustered_vcf1 = os.path.abspath(outprefix + '.1.filter.cluster.vcf')
        self.clustered_vcf2 = os.path.abspath(outprefix + '.2.filter.cluster.vcf')
        self.seqs_out_vcf1 = os.path.abspath(outprefix + '.vcf1.fa')
        self.seqs_out_vcf2 = os.path.abspath(outprefix + '.vcf2.fa')
        self.sam_summary = os.path.abspath(outprefix + '.summary.tsv')
        self.stats_out = os.path.abspath(outprefix + '.stats.tsv')
        self.gt_conf_hist_out = os.path.abspath(outprefix + '.gt_conf_hist.tsv')

        self.flank_length = flank_length
        self.merge_length = flank_length if merge_length is None else merge_length
        self.filter_and_cluster_vcf = filter_and_cluster_vcf
        self.discard_ref_calls = discard_ref_calls
        self.allow_flank_mismatches = allow_flank_mismatches

        if self.filter_and_cluster_vcf:
            self.vcf_to_check1 = self.clustered_vcf1
            self.vcf_to_check2 = self.clustered_vcf2
        else:
            self.vcf_to_check1 = self.vcf_file_in1
            self.vcf_to_check2 = self.vcf_file_in2

        self.exclude_regions1 = DnadiffMappingBasedVerifier._load_exclude_regions_bed_file(exclude_regions_bed_file1)
        self.exclude_regions2 = DnadiffMappingBasedVerifier._load_exclude_regions_bed_file(exclude_regions_bed_file2)
        self.max_soft_clipped = max_soft_clipped
        self.number_ns = 1

    @classmethod
    def _write_dnadiff_plus_flanks_to_fastas(cls, dnadiff_file, ref_infile, query_infile, ref_outfile, query_outfile, flank_length):
        '''Given a dnadiff snps file and the corresponding ref and query fasta,
        write a fasta for each infile containing each variant plus flank_length
        nucleotides added to its start and end.
        Calls each sequence:
            dnadiff_snps_index.start_position.'''
        seq1 = ""
        with open(ref_infile, "r") as in_handle1:
            for record in SeqIO.parse(in_handle1, "fasta"):
                seq1 = str(record.seq)
                break

        seq2 = ""
        with open(query_infile, "r") as in_handle2:
            for record in SeqIO.parse(in_handle2, "fasta"):
                seq2 = str(record.seq)
                break

        out_handle1 = open(ref_outfile, "w")
        out_handle2 = open(query_outfile, "w")

        snps = pd.read_csv(dnadiff_file, sep='\t', header=None)

        for line in snps.itertuples():
            assert(len(line) > 4)
            seq_name = str(line[0]) + "." + str(line[1])
            flanked_seq = ""
            if line[2] == '.':
                start = max(0,line[1] - flank_length)
                end = min(line[1] + flank_length, len(seq1))
                flanked_seq = seq1[start:end]
            else:
                start = max(0, line[1] - flank_length - 1)
                end = min(line[1] + flank_length, len(seq1))
                flanked_seq = seq1[start:end]
            print('>' + seq_name, flanked_seq, sep='\n', file=out_handle1)
            if line[3] == '.':
                start = max(0, line[4] - flank_length)
                end = min(line[4] + flank_length, len(seq2))
                flanked_seq = seq2[start:end]
            else:
                start = max(0, line[4] - flank_length - 1)
                end = min(line[4] + flank_length, len(seq2))
                flanked_seq = seq2[start:end]
            print('>' + seq_name, flanked_seq, sep='\n', file=out_handle2)

        out_handle1.close()
        out_handle2.close()

    @classmethod
    def _load_exclude_regions_bed_file(cls, infile):
        regions = {}
        if infile is not None:
            with open(infile) as f:
                for line in f:
                    fields = line.rstrip().split('\t')
                    if fields[0] not in regions:
                        regions[fields[0]] = []
                    start = int(fields[1])
                    end = int(fields[2]) - 1
                    regions[fields[0]].append(pyfastaq.intervals.Interval(start, end))

            for ref_name in regions:
                pyfastaq.intervals.merge_overlapping_in_list(regions[ref_name])

        return regions

    @classmethod
    def _interval_intersects_an_interval_in_list(cls, interval, interval_list):
        # This could be faster by doing something like a binary search.
        # But we're looking for points in intervals, so fiddly to implement.
        # Not expecting a log interval list, so just do a simple check
        # from start to end for now
        i = 0
        while i < len(interval_list) and interval.start > interval_list[i].end:
            i += 1

        return i < len(interval_list) and interval.intersects(interval_list[i])

    @classmethod
    def _filter_vcf_for_clustering(cls, infile, outfile, discard_ref_calls=True):
        header_lines, vcf_records = vcf_file_read.vcf_file_to_dict(infile, sort=True, homozygous_only=False, remove_asterisk_alts=True, remove_useless_start_nucleotides=True)

        with open(outfile, 'w') as f:
            print(*header_lines, sep='\n', file=f)
            for ref_name in vcf_records:
                for vcf_record in vcf_records[ref_name]:
                    if 'MISMAPPED_UNPLACEABLE' in vcf_record.FILTER:
                        continue
                    if vcf_record.FORMAT is None or 'GT' not in vcf_record.FORMAT:
                        logging.warning('No GT in vcf record:' + str(vcf_record))
                        continue
                    if vcf_record.REF in [".", ""]:
                        continue

                    genotype = vcf_record.FORMAT['GT']
                    genotypes = genotype.split('/')
                    called_alleles = set(genotypes)
                    if len(called_alleles) != 1 or (discard_ref_calls and called_alleles == {'0'}) or '.' in called_alleles:
                        continue

                    if len(vcf_record.ALT) > 1:
                        if called_alleles != {'0'}:
                            vcf_record.set_format_key_value('GT', '1/1')
                            try:
                                vcf_record.ALT = [vcf_record.ALT[int(genotypes[0]) - 1]]
                            except:
                                raise Error('BAD VCf line:' + str(vcf_record))
                        else:
                            vcf_record.set_format_key_value('GT', '0/0')
                            vcf_record.ALT = [vcf_record.ALT[0]]
                    if vcf_record.ALT[0] in [".",""]:
                        continue

                    if vcf_record.FORMAT['GT'] == '0':
                        vcf_record.FORMAT['GT'] = '0/0'
                    elif vcf_record.FORMAT['GT'] == '1':
                        vcf_record.FORMAT['GT'] = '1/1'

                    if 'GL' in vcf_record.FORMAT.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        likelihoods = vcf_record.FORMAT['GL'].split(',')
                        assert(len(likelihoods) > 2)
                        if called_alleles == {'0'}:
                            vcf_record.set_format_key_value('GT_CONF',str(float(likelihoods[0]) - float(likelihoods[1])))
                        else:
                            vcf_record.set_format_key_value('GT_CONF', str(float(likelihoods[int(genotypes[0])]) - float(likelihoods[0])))
                    if 'SupportFraction' in vcf_record.INFO.keys() and 'GT_CONF' not in vcf_record.FORMAT.keys():
                        vcf_record.set_format_key_value('GT_CONF',
                                                        str(float(vcf_record.INFO['SupportFraction'])*100))
                    print(vcf_record, file=f)


    @classmethod
    def _write_vars_plus_flanks_to_fasta(cls, outfile, vcf_records, ref_seqs, flank_length, number_ns=0):
        '''Given a dict of vcf records made by vcf_file_read.vcf_file_to_dict(),
        and its correcsponding file of reference sequences, writes a new fasta file
        of each ref seq and inferred variant sequence plus flank_length nucleotides added to
        its start and end. Calls each sequence:
            ref_name.start_position.vcf_list_index.allele_number
        where allele_numbers in same order as VCF, with ref seq = allele 0.'''
        prev_ref_name = None
        prev_ref_pos = None
        j = 0
        with open(outfile, 'w') as f:
            for ref_name in sorted(vcf_records):
                for i, vcf_record in enumerate(vcf_records[ref_name]):
                    start_position, alleles = vcf_record.inferred_var_seqs_plus_flanks(ref_seqs[ref_name], flank_length)

                    for allele_index, allele_seq in enumerate(alleles):
                        seq_name = '.'.join([ref_name, str(start_position + 1), str(j), str(i), str(allele_index)])
                        allele_seq = allele_seq.replace('.','')
                        for n in range(number_ns):
                            allele_seq = "N" + allele_seq + "N"
                        print('>' + seq_name, allele_seq, sep='\n', file=f)
                    if prev_ref_name == ref_name and prev_ref_pos == start_position:
                        j += 1
                    else:
                        j = 0
                    prev_ref_name = ref_name
                    prev_ref_pos = start_position



    @classmethod
    def _map_seqs_to_seqs(cls, seqs_file_ref, seqs_file_query, outfile):
        '''Map seqs_file to ref_file using BWA MEM.
        Output is SAM file written to outfile'''
        bwa_binary = dependencies.find_binary('bwa')
        command = ' '.join([
            bwa_binary, 'index',
            seqs_file_ref,
        ])
        utils.syscall(command)

        command = ' '.join([
            bwa_binary, 'aln',
            seqs_file_ref,
            seqs_file_query,
            '>', outfile + ".tmp",
        ])
        utils.syscall(command)

        command = ' '.join([
            bwa_binary, 'samse',
            seqs_file_ref,
            outfile + ".tmp",
            seqs_file_query,
            '>', outfile,
        ])
        utils.syscall(command)
        #os.unlink(outfile + ".tmp")

    @classmethod
    def _check_if_sam_match_is_good(cls, sam_record, ref_seqs, flank_length, query_sequence=None, allow_mismatches=True, max_soft_clipped=3):
        logging.debug(f'Checking SAM: {sam_record}')
        if sam_record.is_unmapped:
            return 'Unmapped'

        if not allow_mismatches:
            try:
                nm = sam_record.get_tag('NM')
            except:
                raise Error('No NM tag found in sam record:' + str(sam_record))

            all_mapped = len(sam_record.cigartuples) == 1 and sam_record.cigartuples[0][0] == 0
            if all_mapped and nm == 0:
                logging.debug('SAM record passed no mismatches allowed check')
                return 'Good'
            else:
                logging.debug('SAM record failed no mismatches allowed check')
                return 'Bad_mismatches'

        # don't allow too many soft clipped bases
        if (sam_record.cigartuples[0][0] == 4 and sam_record.cigartuples[0][1] > max_soft_clipped) \
                or (sam_record.cigartuples[-1][0] == 4 and sam_record.cigartuples[-1][1] > max_soft_clipped):
            logging.debug('SAM record failed soft clipping check')
            return 'Bad_soft_clipped'

        if query_sequence is None:
            query_sequence = sam_record.query_sequence
        assert query_sequence is not None

        # if the query is short, which happens when the variant we
        # are checking is too near the start or end of the ref sequence
        if len(query_sequence) < 2 * flank_length + 1:
            # This is an edge case. We don't really know which part
            # of the query seq we're looking for, so guess
            length_diff = 2 * flank_length - len(query_sequence)

            if sam_record.query_alignment_start < 5:
                alt_seq_end = len(query_sequence) - flank_length - 1
                alt_seq_start = min(alt_seq_end, flank_length - length_diff)
            else:
                alt_seq_start = flank_length
                alt_seq_end = max(alt_seq_start, length_diff + len(query_sequence) - flank_length - 1)
        else:
            alt_seq_start = flank_length
            alt_seq_end = len(query_sequence) - flank_length - 1

        aligned_pairs = sam_record.get_aligned_pairs(with_seq=True)
        logging.debug(f'aligned_pairs: {aligned_pairs}')
        wanted_aligned_pairs = []
        current_pos = 0

        i = 0
        while i < len(query_sequence):
            if aligned_pairs[i][0] is None:
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])
            elif current_pos > alt_seq_end:
                break
            else:
                current_pos = aligned_pairs[i][0]
                if alt_seq_start - 1 <= current_pos <= alt_seq_end + 1:
                    wanted_aligned_pairs.append(aligned_pairs[i])

            i += 1

        logging.debug(f'wanted_aligned_pairs: {wanted_aligned_pairs}')
        assert len(wanted_aligned_pairs) > 0

        for pair in wanted_aligned_pairs:
            if None in pair or query_sequence[pair[0]].upper() != pair[2].upper():
                logging.debug('SAM record failed because mismatch in allele sequence plus 1bp either side')
                return 'Bad_allele_mismatch'

        logging.debug('SAM record passed all checks')
        return 'Good'

    @classmethod
    def _index_vcf(cls, vcffile):
        '''Index VCF file'''
        bgzip_binary = dependencies.find_binary('bgzip')
        command = ' '.join([
            bgzip_binary,
            '-c',
            vcffile,
            '>',
            vcffile + ".gz",
        ])
        utils.syscall(command)

        tabix_binary = dependencies.find_binary('tabix')
        command = ' '.join([
            tabix_binary,
            '-p',
            'vcf',
            vcffile + ".gz",
        ])
        utils.syscall(command)

    @classmethod
    def _parse_sam_file_and_vcf(cls, samfile, vcffile, dnadiff_plus_flanks_file, flank_length, allow_mismatches, exclude_regions=None, max_soft_clipped=3, number_ns=0):
        if  exclude_regions is None:
            exclude_regions = {}

        found = []
        match_flag = []
        correct_allele = []
        gt_conf = []
        allele = []
        dnadiff_file_seqs = {}
        pyfastaq.tasks.file_to_dict(dnadiff_plus_flanks_file, dnadiff_file_seqs)
        samfile_handle = pysam.AlignmentFile(samfile, "r")
        sam_previous_record_name = None
        for sam_record in samfile_handle.fetch(until_eof=True):
            if sam_record.query_name == sam_previous_record_name:
                continue
            sam_previous_record_name = sam_record.query_name
            found_conf = False
            found_allele = False

            # see if excluded region in bed file
            var_num, start = sam_record.query_name.rsplit('.', maxsplit=2)
            exclude = False
            for ref_name in exclude_regions.keys():
                end = int(start) + 1
                interval = pyfastaq.intervals.Interval(start, end)
                exclude = DnadiffMappingBasedVerifier._interval_intersects_an_interval_in_list(interval,
                                                                                        exclude_regions[ref_name])
            if exclude:
                found.append('Exclude')
                gt_conf.append(0)
                allele.append('0')
                continue

            match = DnadiffMappingBasedVerifier._check_if_sam_match_is_good(sam_record,
                                                                                 dnadiff_file_seqs,
                                                                                 flank_length,
                                                                                 query_sequence=sam_record.query_sequence,
                                                                                 allow_mismatches=allow_mismatches,
                                                                                 max_soft_clipped=max_soft_clipped)
            alignment_start = str(sam_record).split("\t")[3]
            match_flag.append(match)
            if match == 'Good':
                logging.debug('SAM record is a good match')
                logging.debug('SAM record reference is %s' %sam_record.reference_name)
                ref_name, expected_start, vcf_pos_index, vcf_record_index, allele_index = sam_record.reference_name.rsplit('.', maxsplit=4)

                vcf_reader = pysam.VariantFile(vcffile)
                vcf_interval_start = int(expected_start) + int(alignment_start) + flank_length - 2 - number_ns
                vcf_interval_end = int(expected_start) + int(alignment_start) + flank_length - number_ns
                logging.debug('Find VCF records matching ref %s in interval [%i,%i]' % (ref_name, vcf_interval_start, vcf_interval_end))
                for i, vcf_record in enumerate(vcf_reader.fetch(ref_name, vcf_interval_start, vcf_interval_end)):
                    if i == int(vcf_pos_index):
                        sample_name = vcf_record.samples.keys()[0]
                        if 'GT' in vcf_record.format.keys() and len(set(vcf_record.samples[sample_name]['GT'])) == 1:
                            if int(allele_index) == vcf_record.samples[sample_name]['GT'][0]:
                                found.append('1')
                                allele.append(str(allele_index))
                                correct_allele.append('1')
                                found_allele = True
                                if 'GT_CONF' in vcf_record.format.keys():
                                    gt_conf.append(int(float(vcf_record.samples[sample_name]['GT_CONF'])))
                                    found_conf = True
            if not found_allele:
                found.append('0')
                allele.append('0')
                correct_allele.append('0')
            if not found_conf:
                gt_conf.append(0)
        assert len(found) == len(gt_conf)
        assert len(found) == len(allele)
        assert len(found) == len(match_flag)
        assert len(found) == len(correct_allele)
        return found, gt_conf, allele, match_flag, correct_allele

    @classmethod
    def _parse_sam_files(cls, dnadiff_file, samfile1, samfile2, vcffile1, vcffile2, reffasta1, reffasta2, outfile, flank_length, allow_mismatches=True, exclude_regions1=None, exclude_regions2=None, max_soft_clipped=3, number_ns=0):
        '''Input is the original dnadiff snps file of sites we are searching for
        and 2 SAM files made by _map_seqs_to_seqs(), which show mappings of snp sites
        from from the dnadiff snps file to the vcf (i.e. searches if VCF contains an record
        with the appropriate sequence.
        Creates a tsv detailing whether the snp difference could be detected and at what
        GT_CONF threshold.
        '''

        snps = pd.read_csv(dnadiff_file, sep='\t', header=None)
        ref_found, ref_conf, ref_allele, ref_match_flag, ref_allele_flag = DnadiffMappingBasedVerifier._parse_sam_file_and_vcf(samfile1, vcffile1, reffasta1, flank_length, allow_mismatches, exclude_regions1, max_soft_clipped, number_ns)
        query_found, query_conf, query_allele, query_match_flag, query_allele_flag = DnadiffMappingBasedVerifier._parse_sam_file_and_vcf(samfile2, vcffile2, reffasta2, flank_length, allow_mismatches, exclude_regions2, max_soft_clipped, number_ns)
        logging.debug(f'Length of SNPs to verify: {len(snps[0])}')
        logging.debug(f'Length of ref info found: {len(ref_found)}')
        logging.debug(f'Length of query info found: {len(query_found)}')
        assert len(snps[0]) == len(ref_found)
        assert len(snps[0]) == len(query_found)
        out_df = pd.DataFrame({'id': snps[0],
                               'ref': snps[1],
                               'alt': snps[2],
                               'ref_found': ref_found,
                               'ref_conf' : ref_conf,
                               'ref_allele': ref_allele,
                               'ref_match_flag': ref_match_flag,
                               'ref_allele_correct': ref_allele_flag,
                               'query_found': query_found,
                               'query_conf': query_conf,
                               'query_allele': query_allele,
                               'query_match_flag': query_match_flag,
                               'query_allele_correct': query_allele_flag
                               })
        out_df.to_csv(outfile, sep='\t')

    @classmethod
    def _gather_stats(cls, tsv_file):
        stats = {x: 0 for x in ['total', 'found_vars', 'missed_vars', 'excluded_vars']}
        gt_conf_hist = {}

        snps = pd.read_csv(tsv_file, sep='\t', index_col=0)
        for index,line in snps.iterrows():
            stats['total'] += 1
            if (line['ref_found'] == 'Exclude' or line['query_found'] == 'Exclude'):
                stats['excluded_vars'] += 1
            elif (line['ref_found'] == 1 or line['query_found'] == 1 or line['ref_found'] == '1' or line['query_found'] == '1'):
                stats['found_vars'] += 1
                gt_confs = [i for i in {line['ref_conf'],line['query_conf']} if not math.isnan(i)]
                gt_conf = None
                if len(gt_confs) > 0:
                    gt_conf = max(gt_confs)
                gt_conf_hist[gt_conf] = gt_conf_hist.get(gt_conf, 0) + 1
            else:
                stats['missed_vars'] += 1
        return stats, gt_conf_hist

    def run(self):
        # Write files of sequences to search for in each vcf
        DnadiffMappingBasedVerifier._write_dnadiff_plus_flanks_to_fastas(self.dnadiff_snps_file, self.dnadiff_file1, self.dnadiff_file2, self.seqs_out_dnadiff1, self.seqs_out_dnadiff2, self.flank_length)

        # Cluster together variants in each vcf
        if self.filter_and_cluster_vcf:
            DnadiffMappingBasedVerifier._filter_vcf_for_clustering(self.vcf_file_in1, self.filtered_vcf1, self.discard_ref_calls)
            DnadiffMappingBasedVerifier._filter_vcf_for_clustering(self.vcf_file_in2, self.filtered_vcf2, self.discard_ref_calls)
            if self.discard_ref_calls:
                clusterer1 = vcf_clusterer.VcfClusterer([self.filtered_vcf1], self.vcf_reference_file, self.clustered_vcf1, merge_method='simple', max_distance_between_variants=self.merge_length)
                clusterer2 = vcf_clusterer.VcfClusterer([self.filtered_vcf2], self.vcf_reference_file, self.clustered_vcf2, merge_method='simple', max_distance_between_variants=self.merge_length)
            else:
                clusterer1 = vcf_clusterer.VcfClusterer([self.filtered_vcf1], self.vcf_reference_file, self.clustered_vcf1, merge_method='gt_aware', max_distance_between_variants=self.merge_length)
                clusterer2 = vcf_clusterer.VcfClusterer([self.filtered_vcf2], self.vcf_reference_file, self.clustered_vcf2, merge_method='gt_aware', max_distance_between_variants=self.merge_length)
            clusterer1.run()
            clusterer2.run()

        vcf_header, vcf_records1 = vcf_file_read.vcf_file_to_dict(self.vcf_to_check1, sort=True, remove_useless_start_nucleotides=True)
        vcf_header, vcf_records2 = vcf_file_read.vcf_file_to_dict(self.vcf_to_check2, sort=True, remove_useless_start_nucleotides=True)
        sample_from_header = vcf_file_read.get_sample_name_from_vcf_header_lines(vcf_header)
        if sample_from_header is None:
            sample_from_header = 'sample'
        vcf_ref_seqs = {}
        pyfastaq.tasks.file_to_dict(self.vcf_reference_file, vcf_ref_seqs)

        DnadiffMappingBasedVerifier._write_vars_plus_flanks_to_fasta(self.seqs_out_vcf1, vcf_records1, vcf_ref_seqs, self.flank_length, self.number_ns)
        DnadiffMappingBasedVerifier._write_vars_plus_flanks_to_fasta(self.seqs_out_vcf2, vcf_records2, vcf_ref_seqs, self.flank_length, self.number_ns)
        DnadiffMappingBasedVerifier._map_seqs_to_seqs(self.seqs_out_vcf1, self.seqs_out_dnadiff1, self.sam_file_out1)
        DnadiffMappingBasedVerifier._map_seqs_to_seqs(self.seqs_out_vcf2, self.seqs_out_dnadiff2, self.sam_file_out2)
        #for f in glob.glob(self.seqs_out_vcf1 + '*'):
            #os.unlink(f)
        #for f in glob.glob(self.seqs_out_vcf2 + '*'):
            #os.unlink(f)

        DnadiffMappingBasedVerifier._index_vcf(self.vcf_to_check1)
        self.vcf_to_check1 = self.vcf_to_check1 + ".gz"
        DnadiffMappingBasedVerifier._index_vcf(self.vcf_to_check2)
        self.vcf_to_check2 = self.vcf_to_check2 + ".gz"
        DnadiffMappingBasedVerifier._parse_sam_files(self.dnadiff_snps_file, self.sam_file_out1, self.sam_file_out2,
                                                     self.vcf_to_check1, self.vcf_to_check2, self.seqs_out_dnadiff1,
                                                     self.seqs_out_dnadiff2, self.sam_summary, self.flank_length,
                                                     allow_mismatches=self.allow_flank_mismatches,
                                                     exclude_regions1=self.exclude_regions1,
                                                     exclude_regions2=self.exclude_regions2,
                                                     max_soft_clipped=self.max_soft_clipped,
                                                     number_ns=self.number_ns)
        stats, gt_conf_hist = DnadiffMappingBasedVerifier._gather_stats(self.sam_summary)
        #os.unlink(self.seqs_out_dnadiff1)
        #os.unlink(self.seqs_out_dnadiff2)
        #for f in glob.glob(self.vcf_to_check1 + '*'):
        #    os.unlink(f)
        #for f in glob.glob(self.vcf_to_check2 + '*'):
        #    os.unlink(f)

        # write stats file
        with open(self.stats_out, 'w') as f:
            keys = stats.keys()
            print(*keys, sep='\t', file=f)
            print(*[stats[x] for x in keys], sep='\t', file=f)


        # write GT_CONF histogram files
        with open(self.gt_conf_hist_out, 'w') as f:
            print('GT_CONF\tCount', file=f)
            for gt_conf, count in sorted(gt_conf_hist.items()):
                print(gt_conf, count, sep='\t', file=f)



import copy
from operator import attrgetter
import os

import pyfastaq
import pymummer
from cluster_vcf_records import vcf_clusterer, vcf_file_read, vcf_record

from minos import utils

class Error (Exception): pass

dnadiff_output_extensions = [
    '1coords',
    '1delta',
    'delta',
    'mcoords',
    'mdelta',
    'qdiff',
    'rdiff',
    'report',
    'snps',
    'unqry',
    'unref',
]

class Dnadiff:
    def __init__(self, ref_fasta, query_fasta, outprefix):
        self.ref_fasta = ref_fasta
        self.query_fasta = query_fasta
        self.outprefix = outprefix
        self.ref_seq_names, self.ref_seqs = Dnadiff._load_seq_file(self.ref_fasta)
        self.query_seq_names, self.query_seqs = Dnadiff._load_seq_file(self.query_fasta)


    @classmethod
    def clean_dnadiff_files(cls, prefix):
        for extension in dnadiff_output_extensions:
            # not all files get written, hence try except pass
            try:
                os.unlink(prefix + '.' + extension)
            except:
                pass


    @classmethod
    def _load_seq_file(cls, infile):
        '''Returns tuple of:
        1. list of seqnames in same order as in file, and
        2. dictionary of seqs'''
        seqreader = pyfastaq.sequences.file_reader(infile)
        seqs = {}
        seqnames = []
        for seq in seqreader:
            seqnames.append(seq.id)
            seqs[seq.id] = copy.copy(seq)

        return seqnames, seqs


    @classmethod
    def _run_dnadiff(cls, ref_fasta, query_fasta, outprefix):
        command = ' '.join(['dnadiff -p', outprefix, ref_fasta, query_fasta])
        utils.syscall(command)


    @classmethod
    def _load_qdiff_file(cls, qdiff_file):
        '''Loads the out.qdiff file made by dnadiff.
        Returns dict of seq name -> list of sorted non-overlapping
        regions that contain variants compared to the ref'''
        diff_regions = {}

        with open(qdiff_file) as f:
            for line in f:
                fields = line.rstrip().split()
                start, end = sorted([int(fields[2]) - 1, int(fields[3]) - 1])
                interval = pyfastaq.intervals.Interval(start, end)
                if fields[0] not in diff_regions:
                    diff_regions[fields[0]] = []
                diff_regions[fields[0]].append(interval)

        for interval_list in diff_regions.values():
            pyfastaq.intervals.merge_overlapping_in_list(interval_list)

        return diff_regions


    @classmethod
    def _snps_file_file_to_unmerged_vcf(cls, snps_file, query_seqs, outfile):
        '''Loads the .snps file made by dnadiff.
        query_seqs = dictionary of the query sequences.
        ref_seqs, qry_seqs = dictionaries of the genome sequences.
        Writes a new VCF file unmerged records.'''
        vcf_records = {}
        variants = pymummer.snp_file.get_all_variants(snps_file)

        for variant in variants:
            # If the variant is reversed, it means that either the ref or query had to be
            # reverse complemented when aligned by mummer. Need to do the appropriate
            # reverse (complement) fixes so the VCF has the correct REF and ALT sequences
            if variant.reverse:
                qry_seq = pyfastaq.sequences.Fasta('x', variant.qry_base)
                qry_seq.revcomp()
                variant.qry_base = ''.join(reversed(qry_seq.seq))
                ref_seq = pyfastaq.sequences.Fasta('x', variant.ref_base)
                ref_seq.revcomp()
                variant.ref_base = ref_seq.seq

            if variant.var_type == pymummer.variant.SNP:
                new_record = vcf_record.VcfRecord('\t'.join([
                    variant.qry_name,
                    str(variant.qry_start + 1),
                    '.',
                    variant.qry_base,
                    variant.ref_base,
                    '.',
                    '.',
                    'SVTYPE=DNADIFF_SNP',
                    'GT',
                    '1/1',
                ]))
            elif variant.var_type == pymummer.variant.DEL:
                # The query has sequence missing, compared to the
                # reference. We're making VCF records w.r.t. the
                # query, so this is an insertion. So need to
                # get the nucleotide before the insertion as well.
                new_record = vcf_record.VcfRecord('\t'.join([
                    variant.qry_name,
                    str(variant.qry_start + 1),
                    '.',
                    query_seqs[variant.qry_name][variant.qry_start],
                    query_seqs[variant.qry_name][variant.qry_start] + variant.ref_base,
                    '.',
                    '.',
                    'SVTYPE=DNADIFF_INS',
                    'GT',
                    '1/1',
                ]))
            elif variant.var_type == pymummer.variant.INS:
                # The ref has sequence missing, compared to the
                # query. We're making VCF records w.r.t. the
                # query, so this is a deletion. So need to
                # get the nucleotide before the deletion as well.
                new_record = vcf_record.VcfRecord('\t'.join([
                    variant.qry_name,
                    str(variant.qry_start),
                    '.',
                    query_seqs[variant.qry_name][variant.qry_start - 1] + variant.qry_base,
                    query_seqs[variant.qry_name][variant.qry_start - 1],
                    '.',
                    '.',
                    'SVTYPE=DNADIFF_DEL',
                    'GT',
                    '1/1',
                ]))
            else:
                raise Error('Unknown variant type: ' + str(variant))

            assert new_record.REF == query_seqs[new_record.CHROM][new_record.POS:new_record.POS + len(new_record.REF)]

            if new_record.CHROM not in vcf_records:
                vcf_records[new_record.CHROM] = []

            vcf_records[new_record.CHROM].append(new_record)

        for vcf_list in vcf_records.values():
            vcf_list.sort(key=attrgetter('POS'))

        with open(outfile, 'w') as f:
            for key, vcf_list in sorted(vcf_records.items()):
                for record in vcf_list:
                    print(record, file=f)


    @classmethod
    def _make_all_variants_intervals(cls, variants, big_variant_intervals):
        '''Makes union of all positions where there are variants,
        by combining output of _load_qdiff_file() and variants from merged VCF file'''
        all_ref_seqs = set(variants.keys()).union(set(big_variant_intervals.keys()))
        intervals = {x: [] for x in all_ref_seqs}

        for ref_seq in intervals:
            if ref_seq in variants:
                for variant in variants[ref_seq]:
                    if len(variant.REF) > 1 or len(variant.ALT) > 1:
                        intervals[ref_seq].append(pyfastaq.intervals.Interval(variant.POS + 1, variant.ref_end_pos()))
                    else:
                        intervals[ref_seq].append(pyfastaq.intervals.Interval(variant.POS, variant.ref_end_pos()))

            if ref_seq in big_variant_intervals:
                intervals[ref_seq].extend(big_variant_intervals[ref_seq])

            pyfastaq.intervals.merge_overlapping_in_list(intervals[ref_seq])

        return intervals


    def run(self):
        snps_file = self.outprefix + '.snps'
        qdiff_file = self.outprefix + '.qdiff'
        self.unmerged_vcf = self.outprefix + '.raw.vcf'
        self.merged_vcf = self.outprefix + '.merged.vcf'
        for filename in [snps_file, qdiff_file]:
            if os.path.exists(filename):
                os.unlink(filename)
        tmp_prefix = self.outprefix + '.tmp'

        for ref_name, query_name in zip(self.ref_seq_names, self.query_seq_names):
            ref_fasta = tmp_prefix + '.ref.fa'
            query_fasta = tmp_prefix + '.query.fa'
            with open(ref_fasta, 'w') as f:
                print(self.ref_seqs[ref_name], file=f)
            with open(query_fasta, 'w') as f:
                print(self.query_seqs[query_name], file=f)
            Dnadiff._run_dnadiff(ref_fasta, query_fasta, tmp_prefix)
            utils.syscall('cat ' + tmp_prefix + '.snps >> ' + snps_file)
            utils.syscall('cat ' + tmp_prefix + '.qdiff >> ' + qdiff_file)
            Dnadiff.clean_dnadiff_files(tmp_prefix)
            os.unlink(ref_fasta)
            os.unlink(query_fasta)

        Dnadiff._snps_file_file_to_unmerged_vcf(self.outprefix + '.snps', self.query_seqs, self.unmerged_vcf)
        clusterer = vcf_clusterer.VcfClusterer([self.unmerged_vcf], self.query_fasta, self.merged_vcf, merge_method='simple')
        clusterer.run()
        header, self.variants = vcf_file_read.vcf_file_to_dict(self.merged_vcf, remove_useless_start_nucleotides=True)
        self.big_variant_intervals = Dnadiff._load_qdiff_file(self.outprefix + '.qdiff')
        self.all_variant_intervals = Dnadiff._make_all_variants_intervals(self.variants, self.big_variant_intervals)




import glob
from setuptools import setup, find_packages


setup(
    name='bio-minos',
    version='0.7.0',
    description='Variant call adjudication',
    packages = find_packages(),
    author='Martin Hunt',
    author_email='mhunt@ebi.ac.uk',
    url='https://github.com/iqbal-lab-org/minos',
    scripts=glob.glob('scripts/*'),
    test_suite='nose.collector',
    tests_require=['nose >= 1.3'],
    install_requires=[
        'biopython',
        'cluster_vcf_records >= 0.10.0',
        'gramtools',
        'matplotlib',
        'pandas',
        'pyfastaq >= 3.14.0',
        'pymummer >= 0.11.0',
        'pysam >= 0.12',
        'scipy >= 1.0.0',
        'seaborn',
    ],
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Programming Language :: Python :: 3 :: Only',
        'License :: OSI Approved :: MIT License',
    ],
)




import json
import logging
import os
import shutil

import pyfastaq

from minos import utils, __version__

class Error (Exception): pass


def find_binary(program, allow_fail=False):
    binary = os.environ.get('MINOS_' + program.upper(), program)
    which_output = shutil.which(binary)
    if which_output is None and not allow_fail:
        raise Error('Error finding ' + program + ' in $PATH. Looked for "' + binary + '"')
    return which_output


def get_version_of_program(program, binary=None, allow_fail=False):
    if binary is None:
        binary = find_binary(program, allow_fail=allow_fail)

    if program == 'bcftools':
        bcftools_process = utils.syscall(binary, allow_fail=True)
        for line in bcftools_process.stderr.split('\n'):
            # example version line:
            # Version: 1.3.1 (using htslib 1.3.1)
            if line.startswith('Version:'):
                try:
                    version = line.rstrip().split(maxsplit=1)[1]
                except:
                    return None
                return version
        return None
    elif program == 'gramtools':
        gramtools_process = utils.syscall(binary + ' --version')
        gramtools_json = json.loads(gramtools_process.stdout)
        return gramtools_json.get('version_number', None)
    elif program == 'bwa':
        # To get version of BWA, need to run it with no options.
        # This returns an error code of 1, which we need to ignore
        bwa_process = utils.syscall(binary, allow_fail=True)
        for line in bwa_process.stderr.split('\n'):
            if line.strip().startswith('Version:'):
                try:
                    version = line.rstrip().split()[-1]
                except:
                    return None
                return version
        return None
    elif program == 'dnadiff':
        dnadiff_process = utils.syscall(binary + ' --version', allow_fail=True)
        for line in dnadiff_process.stderr.split('\n'):
            if line.strip().startswith('DNAdiff version'):
                try:
                    version = line.rstrip().split()[-1]
                except:
                    return None
                return version
    elif program == 'nextflow':
        nextflow_process = utils.syscall(binary + ' -version', allow_fail=True)
        # example line that we want to capture;
        #       version 0.27.6 build 4775
        for line in nextflow_process.stdout.split('\n'):
            if line.strip().startswith('version'):
                try:
                    version = line.rstrip().split(maxsplit=1)[1]
                except:
                    return None
                return version
        return None
    else:
        raise Error('Program name "' + program + '" not recognised. Cannot continue')


def find_python_packages():
    '''Returns a dict of data on required
    third-party python packeges, and minos itself:
    package_name => (version, path).
    Values will be None if package not found'''
    packages = {}
    for package in ['minos', 'pyfastaq', 'pymummer', 'pysam']:
        try:
            exec('import ' + package)
            version = eval(package + '.__version__')
            path = eval(package + '.__file__')
        except:
            version = 'NOT_FOUND'
            path = 'NOT_FOUND'

        packages[package] = (version, path)

    return packages


def find_binaries_and_versions(programs=None):
    '''Returns a dict of data on required binaries:
    package_name => (version, path).
        Values will be None if package not found'''
    data = {}

    if programs is None:
        programs = ['bwa', 'dnadiff', 'gramtools', 'nextflow']

    for program in programs:
        binary = find_binary(program, allow_fail=True)
        if binary is None:
            binary = 'NOT_FOUND'
            version = 'NOT_FOUND'
        else:
            version = get_version_of_program(program, binary, allow_fail=True)
            if version is None:
                version = 'NOT_FOUND'

        data[program] = (version, binary)

    return data


def dependencies_report(programs=None):
    '''Returns tuple:
    bool, list
    bool = True iff all dependencies are found
    list = a list of lines that are a
    report on the third party programs and Python
    packages required by minos'''
    all_ok = True
    lines = ['minos ' + __version__]

    if programs is None:
        programs = ['bwa', 'dnadiff', 'gramtools', 'nextflow']
    programs_data = find_binaries_and_versions(programs=programs)

    for program in programs:
        lines.append(program + ' ' + ' '.join(programs_data[program]))
        if 'NOT_FOUND' in programs_data[program]:
            all_ok = False

    packages_data = find_python_packages()
    for package in sorted(packages_data):
        lines.append(package + ' ' + ' '.join(packages_data[package]))
        if 'NOT_FOUND' in packages_data[package]:
            all_ok = False

    return all_ok, lines


def check_and_report_dependencies(outfile=None, programs=None):
    '''Writes report of depndencies to file (could be stdout).
    Raises error if any depndency not found'''
    all_ok, report_lines = dependencies_report(programs=programs)
    if outfile is not None:
        f = pyfastaq.utils.open_file_write(outfile)
        print(*report_lines, sep='\n', file=f)
        print('All ok:', all_ok, file=f)
        pyfastaq.utils.close(f)
    else:
        for line in report_lines:
            logging.info('Depencency check: ' + line)
    if not all_ok:
        raise Error('At least one dependency not found')




from cluster_vcf_records import vcf_clusterer

def run(options):
    clusterer = vcf_clusterer.VcfClusterer(
        options.vcf_files,
        options.ref_fasta,
        options.outfile,
        max_distance_between_variants=options.max_var_dist,
        max_alleles_per_cluster=options.max_alleles_per_cluster,
    )
    clusterer.run()




from cluster_vcf_records import vcf_file_read

class Error (Exception): pass

class VcfFileSplitDeletions:
    def __init__(self,
        infile,
        outfile_small_vars,
        outfile_long_deletions,
        min_large_ref_length=50,
    ):
        self.infile = infile
        self.outfile_small_vars = outfile_small_vars
        self.outfile_long_deletions = outfile_long_deletions
        self.min_large_ref_length = min_large_ref_length


    def run(self):
        header_lines, vcf_records = vcf_file_read.vcf_file_to_list(self.infile)

        with open(self.outfile_small_vars, 'w') as f_small, open(self.outfile_long_deletions, 'w') as f_big:
            print(*header_lines, sep='\n', file=f_small)
            print(*header_lines, sep='\n', file=f_big)

            for original_record in vcf_records:
                split_records = original_record.split_into_snps()
                for record in split_records:
                    if len(record.REF) < self.min_large_ref_length:
                        print(record, file=f_small)
                    else:
                        print(record, file=f_big)



from minos import evaluate_recall

def run(options):
    verifier = evaluate_recall.EvaluateRecall(
        options.truth_vcf_file,
        options.truth_vcf_ref,
        options.query_vcf_file,
        options.query_vcf_ref,
        options.outprefix,
        flank_length=options.flank_length,
        merge_length=options.variant_merge_length,
        filter_and_cluster_vcf=not options.no_filter_cluster,
        discard_ref_calls=not options.include_ref_calls,
        allow_flank_mismatches=options.allow_flank_mismatches,
        exclude_regions_bed_file=options.exclude_bed,
        max_soft_clipped=options.max_soft_clipped,
    )
    verifier.run()




from collections import namedtuple
import itertools
import logging
import multiprocessing
import os
import pickle

import cluster_vcf_records

from minos import gramtools

class Error (Exception): pass


split_file_attributes = [
    'filename',
    'file_number',
    'chrom',
    'chrom_start',
    'chrom_end',
    'file_start_index',
    'file_end_index',
    'use_start_index',
    'use_end_index',
    'gramtools_build_dir',
]
SplitFile = namedtuple('SplitFile', split_file_attributes)


def _run_gramtools_build(split_file, ref_fasta, max_read_length, kmer_size):
    logging.info('Start gramtools build ' + split_file.filename)
    gramtools.run_gramtools_build(split_file.gramtools_build_dir, split_file.filename, ref_fasta, max_read_length, kmer_size)
    logging.info('Finish gramtools build ' + split_file.filename)


class VcfChunker:
    def __init__(self, outdir, vcf_infile=None, ref_fasta=None, variants_per_split=None, max_read_length=200, total_splits=100, flank_length=200, gramtools_kmer_size=10, alleles_per_split=None, threads=1):
        self.outdir = os.path.abspath(outdir)
        self.metadata_pickle = os.path.join(self.outdir, 'data.pickle')
        self.threads = threads

        if os.path.exists(self.outdir):
            self._load_existing_data()
        else:
            logging.info('Setting up new chunked VCF directory ' + self.outdir)
            assert vcf_infile is not None
            assert ref_fasta is not None
            assert None in {variants_per_split, alleles_per_split}
            self.vcf_infile = os.path.abspath(vcf_infile)
            self.ref_fasta = os.path.abspath(ref_fasta)
            self.variants_per_split = variants_per_split
            self.alleles_per_split = alleles_per_split
            self.total_splits = total_splits
            self.flank_length = flank_length
            self.gramtools_kmer_size = gramtools_kmer_size
            self.max_read_length = max_read_length

            if not os.path.exists(self.vcf_infile):
                raise Error('VCF file not found: ' + self.vcf_infile)
            if not os.path.exists(self.ref_fasta):
                raise Error('Reference FASTA file not found: ' + self.ref_fasta)

            try:
                os.mkdir(self.outdir)
            except:
                raise Error('Error mkdir ' + self.outdir)

            self.vcf_split_files = {} # ref name -> list of SplitFile


    def _save_metadata(self):
        metadata = {
            'vcf_infile': self.vcf_infile,
            'ref_fasta': self.ref_fasta,
            'variants_per_split': self.variants_per_split,
            'alleles_per_split': self.alleles_per_split,
            'total_splits': self.total_splits,
            'flank_length': self.flank_length,
            'gramtools_kmer_size': self.gramtools_kmer_size,
            'max_read_length': self.max_read_length,
            'total_split_files': self.total_split_files,
            'split_files': self.vcf_split_files,
            'total_input_records': self.total_input_records,
        }
        with open(self.metadata_pickle, 'wb') as f:
            pickle.dump(metadata, f, pickle.HIGHEST_PROTOCOL)


    def _load_existing_data(self):
        with open(self.metadata_pickle, 'rb') as f:
            metadata = pickle.load(f)

        self.vcf_infile = metadata['vcf_infile']
        self.ref_fasta = metadata['ref_fasta']
        self.variants_per_split = metadata['variants_per_split']
        self.alleles_per_split = metadata['alleles_per_split']
        self.total_splits = metadata['total_splits']
        self.flank_length = metadata['flank_length']
        self.gramtools_kmer_size = metadata['gramtools_kmer_size']
        self.max_read_length = metadata['max_read_length']
        self.total_split_files = metadata['total_split_files']
        self.vcf_split_files = metadata['split_files']
        self.total_input_records = metadata['total_input_records']
        logging.info('Loaded existing data from chunked VCF directory ' + self.outdir)


    @classmethod
    def _chunk_end_indexes_from_vcf_record_list(cls, record_list, start_index, flank_length, total_sites=None, total_alleles=None):
        '''Returns tuple of:
           1. last index of VCF record that we want to use for variant calling
           2. index of last variant in the chunk, which can't be used for variant calling
              but should end up in the gramtools graph'''
        assert total_sites != total_alleles and (total_sites is not None or total_alleles is not None)
        file_start_index = start_index
        while file_start_index > 0:
            distance_to_previous_variant = record_list[start_index].POS - record_list[file_start_index - 1].ref_end_pos()
            if distance_to_previous_variant > flank_length:
                break
            file_start_index -= 1

        if total_sites is not None:
            use_vcf_end_index = min(start_index + total_sites - 1, len(record_list) - 1)
        else:
            use_vcf_end_index = start_index
            alleles = 1 + len(record_list[start_index].ALT)
            while use_vcf_end_index < len(record_list) - 1 and alleles <= total_alleles:
                use_vcf_end_index += 1
                alleles += 1 + len(record_list[use_vcf_end_index].ALT)
            use_vcf_end_index = max(start_index, use_vcf_end_index - 1)

        if use_vcf_end_index == len(record_list) - 1:
            return file_start_index, use_vcf_end_index, use_vcf_end_index

        file_end_index = use_vcf_end_index

        while file_end_index < len(record_list) - 1:
            distance_to_next_variant = record_list[file_end_index + 1].POS - record_list[use_vcf_end_index].ref_end_pos()
            if distance_to_next_variant > flank_length:
                break
            file_end_index += 1

        return file_start_index, use_vcf_end_index, min(file_end_index, len(record_list) - 1)


    @classmethod
    def _total_variants_and_alleles_in_vcf_dict(cls, vcf_dict):
        total_variants = 0
        total_alleles = 0
        for ref, vcf_list in vcf_dict.items():
            total_variants += len(vcf_list)
            total_alleles += sum([1 + len(x.ALT) for x in vcf_list])
        return total_variants, total_alleles


    def make_split_vcf_files(self):
        if len(self.vcf_split_files) > 0:
            return

        self.total_split_files = 0
        self.total_input_records = 0
        vcf_header_lines, vcf_records = cluster_vcf_records.vcf_file_read.vcf_file_to_dict(self.vcf_infile)
        if self.variants_per_split is None and self.alleles_per_split is None:
            total_records, total_alleles = VcfChunker._total_variants_and_alleles_in_vcf_dict(vcf_records)
            self.alleles_per_split = 1 + int(total_alleles / self.total_splits)

        for ref_name, vcf_record_list in vcf_records.items():
            file_end_index = -1
            use_end_index = -1
            self.vcf_split_files[ref_name] = []
            self.total_input_records += len(vcf_record_list)

            while use_end_index < len(vcf_record_list) - 1:
                if file_end_index == -1:
                    use_start_index = 0
                else:
                    use_start_index = self.vcf_split_files[ref_name][-1].use_end_index + 1

                file_start_index, use_end_index, file_end_index = VcfChunker._chunk_end_indexes_from_vcf_record_list(vcf_record_list, use_start_index, self.flank_length, total_sites=self.variants_per_split, total_alleles=self.alleles_per_split)
                split_file = SplitFile(
                    os.path.join(self.outdir, 'split.' + str(self.total_split_files) + '.in.vcf'),
                    self.total_split_files,
                    ref_name,
                    max(0, min(vcf_record_list[file_start_index].POS, vcf_record_list[use_start_index].POS - self.flank_length)),
                    max(vcf_record_list[file_end_index].ref_end_pos(), vcf_record_list[use_end_index].ref_end_pos() + self.flank_length),
                    file_start_index,
                    file_end_index,
                    use_start_index,
                    use_end_index,
                    os.path.join(self.outdir, 'split.' + str(self.total_split_files) + '.gramtools_build'),
                )

                self.vcf_split_files[ref_name].append(split_file)

                with open(split_file.filename, 'w') as f:
                    print(*vcf_header_lines, sep='\n', file=f)
                    for i in range(file_start_index, file_end_index + 1, 1):
                        print(vcf_record_list[i], file=f)

                self.total_split_files += 1
                logging.info('Made split VCF file ' + split_file.filename + '. Total split files: ' + str(self.total_split_files))

        self._save_metadata()


    def run_gramtools_build_on_each_split(self):
        if self.threads == 1:
            for file_list in self.vcf_split_files.values():
                for split_file in file_list:
                    gramtools.run_gramtools_build(split_file.gramtools_build_dir, split_file.filename, self.ref_fasta, self.max_read_length, self.gramtools_kmer_size)
        else:
            assert self.threads > 1
            split_files = []
            for file_list in self.vcf_split_files.values():
                split_files.extend(file_list)

            pool = multiprocessing.Pool(self.threads)
            pool.starmap(_run_gramtools_build, zip(file_list, itertools.repeat(self.ref_fasta), itertools.repeat(self.max_read_length), itertools.repeat(self.gramtools_kmer_size)))
            pool.close()
            pool.join()


    def make_split_files(self):
        if len(self.vcf_split_files) > 0:
            return

        self.make_split_vcf_files()
        self.run_gramtools_build_on_each_split()


    def merge_files(self, files_to_merge, outfile):
        total_output_records = 0
        printed_header_lines = False
        logging.info('Making merged VCF file ' + outfile)

        with open(outfile, 'w') as f:
            for ref_name in self.vcf_split_files:
                assert ref_name in files_to_merge
                assert len(self.vcf_split_files[ref_name]) == len(files_to_merge[ref_name])
                for i, split_file in enumerate(self.vcf_split_files[ref_name]):
                    header_lines, records_to_merge = cluster_vcf_records.vcf_file_read.vcf_file_to_list(files_to_merge[ref_name][i])
                    if not printed_header_lines:
                        print(*header_lines, sep='\n', file=f)
                        printed_header_lines = True
                    start_i = split_file.use_start_index - split_file.file_start_index
                    end_i = start_i + split_file.use_end_index - split_file.use_start_index
                    for j in range(start_i, end_i + 1, 1):
                        total_output_records += 1
                        print(records_to_merge[j], file=f)

        if self.total_input_records != total_output_records:
            raise Error('Number of input VCF records = ' + str(self.total_input_records) + ' != ' + str(total_output_records) + ' = numnber of output VCF records. Cannot continue')

        logging.info('Finished making merged VCF file. Total records: ' + str(self.total_input_records))
